from __future__ import annotations
from numpy.typing import NDArray
from numpy import arange, float64, zeros, pi, sqrt


def interpolate_abciss(signals: tuple[Signal, ...]) -> NDArray[float64]:
    """
    return the axis that would be used by the interpolate function
    """
    rates: list[float] = [signal.rate for signal in signals]

    start: float = max([signal.x[0] for signal in signals])
    end: float = min([signal.x[-1] for signal in signals])

    return arange(start, end, 1 / max(rates))


def interpolate(signals: tuple[Signal, ...]) -> tuple[Signal, ...]:
    """
    Interpolate multiple signals with a single abciss list, which has
    the smallest interval and the bigget rate
    """
    splines = [signal.spline() for signal in signals]

    x = interpolate_abciss(signals)

    new_signals = [
        Signal(x, spline(x), signals[0].settings) for spline in splines
    ]

    return tuple(new_signals)


def compute_light(
    scatter_factor: list[float],
    coupling: list[Signal],
    movement: Signal,
    wavelength: float,
    power_in: float,
    power_out: float,
) -> Signal:
    """
    Compute the projection from a given coupling and movement
    """
    frequencies = interpolate_abciss(
        (movement.sin().psd().sqrt(), coupling[0].abs())
    )
    parts = zeros(
        (
            len(coupling),
            len(frequencies),
        )
    )

    for index in range(2):
        phase = (index + 1) * 4 * pi / wavelength

        factor_n = (movement * phase).sin().psd().sqrt()
        coupling_n = coupling[0].abs()
        factor_d = (movement * phase).cos().psd().sqrt()
        coupling_d = coupling[1].abs()

        factor_n, coupling_n, factor_d, coupling_d = interpolate(
            (factor_n, coupling_n, factor_d, coupling_d)
        )

        parts[index] = (
            sqrt(scatter_factor[index])
            * power_in
            / power_out
            * (coupling_n * factor_n + coupling_d * factor_d).y
        )

    return Signal(
        frequencies,
        sum(parts),
        movement.settings,
    )


def opt_compute_light(
    scatter_factor: float,
    factor_n: Signal,
    coupling_n: Signal,
    factor_d: Signal,
    coupling_d: Signal,
    power_in: float,
    power_out: float,
) -> NDArray[float64]:
    """
    Optimized computing of light with pre-computed factor
    """
    return (
        sqrt(scatter_factor)
        * power_in
        / power_out
        * (coupling_n * factor_n + coupling_d * factor_d).y
    )


from backscattering_analyzer.signal import Signal  # no circular import
